 /*
 * xferstats, a wu-ftpd and ncftpd logfile parser and report generator
 * Copyright 1997  Phil Schwan <pschwan@apk.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 	v1.01 changes:
		* allowed for > 100kb/sec transfer rates in the totals
		* fixed hostname parsing for hostnames without a '.' (ie,
			localhost).  they now show as unresolved
		* made -A the default
		* made parsing compatible with the new way of documenting guest
			logins ('g' instead of 'a')

	v1.02 changes (Sep 27 1997):
		* added -s <section> support
		* made a Makefile for the compiling-impaired
		* dotted quads now show as "unresolved" (as they should!)

	v1.03 changes (Oct 25 1997):
		* check to make sure the xferlog file actually exists
			thanks to: Emil Isberg <emil.isberg@mds.mdh.se>
		* added a small Solaris compatibility flag
			thanks to: Emil Isberg <emil.isberg@mds.mdh.se>
		* added INSTALL file and man page, modified the Makefile

	v1.10 changes (Nov 06 1997):
		* added the "Most Downloaded Files" section
		* added corresponding -p <length> parameter
		* fixed a typo that only affects DEBUG output, but in a big way
		* added more configurable options and split them off into
			config.h
		* boosted variable sizes to accomodate huge xferlogs
		* fixed dotted quad recognition
		* a big big big thanks to Nat Friedman <ndf@mit.edu> and
			Elliot Lee <sopwith@redhat.com> for taking a look at
			the code and sending a couple key fixes

	v1.15 changes (Nov 16 1997):
		* optional HTML output
		* GIF graphs to go along with the HTML output
		* the domain totals are now sorted (by bytes transferred)
		* added the ability to limit the number of lines in certain
			reports.  added the corresponding config.h entries and
			command line arguments.
		* a new cmdline parser instead of relying on getopt, which has
			annoyed me for a while.  this also eliminates the need
			for a separate solaris entry in the Makefile.
		* accepts the logfiles from stdin, just place a single hyphen
			somewhere in the parameter list
		* xferstats now parses ncftpd-generated logfiles too!

 */

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "config.h"

#ifndef NOGRAPH
	#include "gd/gd.h"
	#include "gd/gdfontt.h"
	#include "gd/gdfonts.h"
	#include "gd/gdfontmb.h"
	#include "gd/gdfontl.h"
	#include "gd/gdfontg.h"
#endif

#define STRCMP(x, y) (printf("strcmp at line:%d\n", __LINE__),my_strcmp((x),(y)))
#define MALLOC(x) (printf("malloc at line:%d\n", __LINE__),my_malloc(x))
#define DODOMAIN(x) (printf("do domain at line: %d\n", __LINE__),doDomain(x))

#ifdef DEBUG
int my_strcmp(char * a, char * b)
{
  int retval;
  printf("strcmp(%p,%p)", a, b);
  retval = strcmp(a, b);
  printf(".done\n");
  return retval;
}

void * my_malloc(size_t s)
{
  void * retval;
  printf("mallocing size %d\n", s);
  retval = malloc(s);
  printf("done\n");
  printf("malloc ptr: %p\n", retval);
  return retval;
}
#endif

static char VERSION[] = "xferstats.c  1.15  Copyright (C) Phil Schwan";

const char *MONTHS[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
	"Aug", "Sep", "Oct", "Nov", "Dec", NULL};

static FILE *filePtr;
static char domain[11], onlyDomain[5] = {0}, section[255] = {0}, date[25],
	userType, host[201], fileName[201]={}, lowDate[25] = {0}, highDate[25]
	= {0}, inOut, *fileStr;
static int depth=4, secs, realTraffic = 1, anonTraffic = 1, hourlyTraffic = 1,
	domainTraffic = 1, totalTraffic = 1, mostDownloads = 1, mostDLNumber =
	NUMBER_MOSTDL_STATS, sectionTraffic = 0, sectionLength = 0,
	outputHTML = DEFAULT_HTML, numberDailyStats = NUMBER_DAILY_STATS,
	numberSectionStats = NUMBER_SECTION_STATS, numberDomainStats =
	NUMBER_DOMAIN_STATS, use_stdin = 0, ftpd = DEFAULT_LOG_TYPE;
static unsigned long long total, data;
static unsigned int fileCount = 0;
struct domains {
	char name[11];
	unsigned long long data;
	unsigned int fileCount;
	int sorted;
	unsigned int seconds;
	struct domains *nextPtr;
};
struct days {
	char name[16];
	unsigned long long data;
	unsigned int fileCount;
	unsigned int seconds;
	struct days *nextPtr;
};
struct dirs {
	char name[1024];
	unsigned long long data;
	unsigned int fileCount;
	int sorted;
	struct dirs *nextPtr;
};
struct hour {
	unsigned long long data;
	unsigned int seconds;
	unsigned int fileCount;
	float hixfer;
};	
struct files {
	char name[1024];
	unsigned long long data;
	unsigned int fileCount;
	int sorted;
	struct files *nextPtr;
};
FILE *gif;
struct hour hourStruct[24];
typedef struct domains DomainNode;
typedef DomainNode *DOMPTR;
typedef struct days DayNode;
typedef DayNode *DAYPTR;
typedef struct dirs DirNode;
typedef DirNode *DIRPTR;
typedef struct files FileNode;
typedef FileNode *FILEPTR;

int parseline(void)
{
	char foo[4096]={}, junkChar, tmpStr[1024];

	if (!feof(filePtr)) {
		fgets(foo, 4096, filePtr);  /* there's probably a better way to do this :) */
#ifdef DEBUG
		printf("RAW: %s", foo);
#endif
		if (ftpd == 0) {
			memset(date, 0, sizeof(date));
			strncpy(date, foo, 24);
			sscanf(foo + 25, "%d %s %llu %s %c %c %c %c", &secs, host, &data, fileName, &junkChar, &junkChar, &inOut, &userType);

#ifdef DEBUG
			printf("%s %d %s %llu %s %c %c\n", date, secs, host, data, fileName, inOut, userType);
#endif
		} else {

		/* this is a fairly ugly quick hack, but it seems to work.  I
		   don't plan on replacing this parsing code util a) ncftpd
		   logfiles change or b) someone sends me better code :) */

			char userName[10], *tmpToken;
			int temp;

#ifdef DEBUG
			printf("RAW: %s", foo);
#endif
			while (!feof(filePtr) && (foo[29] != ',' || (foo[28] != 'R' && foo[28] != 'S'))) {
				fgets(foo, 4096, filePtr);
#ifdef DEBUG
				printf("RAW: %s", foo);
#endif
			}
			memset(date, 0, sizeof(date));
			if ((foo[28] == 'R' || foo[28] == 'S') && foo[29] == ',') {
 				strncpy(date, "    ", 4);
				memset(userName, 0, sizeof(userName));
				strncpy(userName, foo+3, 2);
				temp = atoi(userName);
				strncat(date, MONTHS[temp - 1], 3);
				strcat(date, " ");
				strncat(date, foo+6, 2);
				strcat(date, " ");
				strncat(date, foo+9, 8);
				strcat(date, " 19");
				strncat(date, foo, 2);
				tmpToken = strtok(foo+30, ",");
				strcpy(fileName, tmpToken);
				tmpToken = strtok(NULL, ",");
				strcpy(tmpStr, tmpToken);
				sscanf(tmpStr, "%llu", &data);
				tmpToken = strtok(NULL, ",");
				strcpy(tmpStr, tmpToken);
				sscanf(tmpStr, "%d", &secs);
				tmpToken = strtok(NULL, ",");
				tmpToken = strtok(NULL, ",");
				tmpToken = strtok(NULL, ",");
				tmpToken = strtok(NULL, ",");
				strcpy(host, tmpToken);
				if (foo[28] == 'R') inOut = 'o';
				else inOut = 'i';
				if (!strcmp(userName, "anonymous")) userType = 'a';
				else userType = 'r';
#ifdef DEBUG
				printf("%s %d %s %llu %s %c %c\n", date, secs, host, data, fileName, inOut, userType);
#endif
			}	
		}
	} else return 0;
	if (!strcmp(date, ""))  /* this is a hack to make it realize when */
		return 0; 	/* the EOF doesn't contain a valid line */

	return 1;
} /* parseline */


void doHours(void)
{
	int hours;
	char foo[3] = {0};

	strncpy(foo, date + 11, 2);
	hours = atoi(foo);
	hourStruct[hours].data += data;
	hourStruct[hours].fileCount++;
	hourStruct[hours].seconds += secs;
	if ((float) data / secs / 1024 > hourStruct[hours].hixfer)
		hourStruct[hours].hixfer = (float) data / secs / 1024;
} /* doHours */


int finalDir(DIRPTR *sPtr, char dcat[4096])
{
	DIRPTR newPtr;

	if (*sPtr == NULL) {
		newPtr = malloc(sizeof(DirNode));
		
		if (newPtr != NULL) {  /* is space available? */
			memset(newPtr->name, 0, sizeof(newPtr->name));
			if (outputHTML) strncpy(newPtr->name, dcat, 1023);
			else strncpy(newPtr->name, dcat, 46);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->nextPtr = NULL;
			newPtr->sorted = 0;
			*sPtr = newPtr;
#ifdef DEBUG
			printf("1st directory structure for \"%s\" created.\n", newPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}

	{
		DIRPTR currentPtr = *sPtr;
		char tempDir[1024] = {0};

		if (outputHTML) strncpy(tempDir, dcat, 1023);
		else strncpy(tempDir, dcat, 46);
		while (currentPtr != NULL) {
			if (!strcmp(currentPtr->name, tempDir)) {
				currentPtr->fileCount++;
				currentPtr->data += data;
#ifdef DEBUG
				printf("Added stats to the dir entry \"%s\".\n", currentPtr->name);
#endif
				return 1;
			}
			currentPtr = currentPtr->nextPtr;
		}
		newPtr = malloc(sizeof(DirNode));
	
		if (newPtr != NULL) {			/* is space available? */
	    		memset(newPtr->name, 0, sizeof(newPtr->name));
			if (outputHTML) strncpy(newPtr->name, dcat, 1023);
			else strncpy(newPtr->name, dcat, 46);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->sorted = 0;
			currentPtr = *sPtr;
			while (currentPtr->nextPtr != NULL)
				currentPtr = currentPtr->nextPtr;
			currentPtr->nextPtr = newPtr;
			newPtr->nextPtr = NULL;
#ifdef DEBUG
			printf("New directory structure for \"%s\" created.\n", newPtr->name);
#endif
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}
	return 1;
} /* finalDir */


int finalFile(FILEPTR *fPtr, char fcat[1024])
{
	FILEPTR newFPtr;

	if (*fPtr == NULL) {
		newFPtr = malloc(sizeof(FileNode));
		
		if (newFPtr != NULL) {  /* is space available? */
			memset(newFPtr->name, 0, sizeof(newFPtr->name));
			strncpy(newFPtr->name, fcat, 1023);
			newFPtr->data = data;
			newFPtr->fileCount = 1;
			newFPtr->nextPtr = NULL;
			newFPtr->sorted = 0;
			*fPtr = newFPtr;
#ifdef DEBUG
			printf("1st file structure for \"%s\" created.\n", newFPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}

	{
		FILEPTR currentPtr = *fPtr;

		while (currentPtr != NULL) {
			if (!strcmp(currentPtr->name, fcat)) {
				currentPtr->fileCount++;
				currentPtr->data += data;
#ifdef DEBUG
				printf("Added stats to the file entry \"%s\".\n", currentPtr->name);
#endif
				return 1;
			}
			currentPtr = currentPtr->nextPtr;
		}
		newFPtr = malloc(sizeof(FileNode));
	
		if (newFPtr != NULL) {			/* is space available? */
	    		memset(newFPtr->name, 0, sizeof(newFPtr->name));
			strncpy(newFPtr->name, fcat, 1023);
			newFPtr->data = data;
			newFPtr->fileCount = 1;
			newFPtr->sorted = 0;
			currentPtr = *fPtr;
			while (currentPtr->nextPtr != NULL)
				currentPtr = currentPtr->nextPtr;
			currentPtr->nextPtr = newFPtr;
			newFPtr->nextPtr = NULL;
#ifdef DEBUG
			printf("New file structure for \"%s\" created.\n", newFPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}
	return 1;
} /* finalFile */


int doDirs(DIRPTR *sPtr, FILEPTR *fPtr)
{
	char **tokens = NULL, *tmpToken, dcat[4096], fcat[1024];
	int i = 0, fuck_me = 1;

	tmpToken = strtok(fileName, "/");
	for(; tmpToken; i++) {
		tokens = realloc(tokens, sizeof(char *) * (i + 1));
		tokens[i] = strdup(tmpToken);
		tmpToken = strtok(NULL, "/");
	}

	if(!tokens) {	
#ifdef DEBUG
		printf("Bad file name while splitting in doDirs.\n");
#endif
		return 0;
	}

	bzero(dcat, 4096);
	bzero(fcat, 1024);
	for (; fuck_me <= i; fuck_me++) {
		if (fuck_me < depth && fuck_me < i) {
			strcat(dcat, "/");
			strcat(dcat, tokens[fuck_me - 1]);
		}
		strcat(fcat, "/");
		strcat(fcat, tokens[fuck_me - 1]);
	}

#ifdef DEBUG
	{
		printf("Final dir: \"%s\"\n", dcat);
		printf("Final file: \"%s\"\n", fcat);
	}
#endif

	free(tokens);

	if (mostDownloads && inOut == 'o')
		if (!finalFile(fPtr, fcat)) return 0;
	if (!finalDir(sPtr, dcat)) return 0;
	return 1;
} /* doDirs */


int doDaily(DAYPTR *sPtr)
{
	DAYPTR newPtr;
	
	if (*sPtr == NULL) {
		newPtr = malloc(sizeof(DayNode));
		
		if (newPtr != NULL) {  /* is space available? */
			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, date, 10);
			strncat(newPtr->name, date + 19, 5);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->nextPtr = NULL;
			newPtr->seconds = secs;
			*sPtr = newPtr;
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	} else {
		DAYPTR currentPtr = *sPtr;
		char tempDay[16] = {0};

		strncpy(tempDay, date, 10);
		strncat(tempDay, date + 19, 5);

		while (currentPtr != NULL) {
#ifdef DEBUG
			printf("currPtr->name: %s\n", currentPtr->name);
#endif
			if (!strcmp(currentPtr->name, tempDay)) {
				currentPtr->fileCount++;
				currentPtr->data += data;
				currentPtr->seconds += secs;
#ifdef DEBUG
				printf("Added stats to the day \"%s\".\n", currentPtr->name);
#endif
				return 1;
			}
			currentPtr = currentPtr->nextPtr;
		}

	        newPtr = malloc(sizeof(DayNode));
        
       		if (newPtr != NULL) {  /* is space available? */
       			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, date, 10);
			strncat(newPtr->name, date + 19, 5);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->seconds = secs;
			currentPtr = *sPtr;
			while (currentPtr->nextPtr != NULL)
				currentPtr = currentPtr->nextPtr;
			currentPtr->nextPtr = newPtr;
			newPtr->nextPtr = NULL;
#ifdef DEBUG
			printf("New day structure for \"%s\" created.\n", newPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}	
	return 0;
} /* doDaily */
	

int doDomain(DOMPTR *sPtr)
{
	DOMPTR newPtr;
	
	if (*sPtr == NULL) {
		newPtr = malloc(sizeof(DomainNode));
		
		if (newPtr != NULL) {  /* is space available? */
			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, domain, 10);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->nextPtr = NULL;
			newPtr->seconds = secs;
			*sPtr = newPtr;
#ifdef DEBUG
			printf("1st domain (\"%s\") created.\n", newPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	} else {
		DOMPTR currentPtr = *sPtr;

#ifdef DEBUG
		printf("DOMAIN: %s\n", domain);
#endif
		while (currentPtr != NULL) {
#ifdef DEBUG
 			printf("cPtr  : %s\n", currentPtr->name);
#endif
			if (!strcmp(currentPtr->name, domain)) {
				currentPtr->fileCount++;
				currentPtr->data += data;
				currentPtr->seconds += secs;
#ifdef DEBUG
				printf("Added to existing domain \"%s\".\n", currentPtr->name);
#endif
				return 1;
			}
			currentPtr = currentPtr->nextPtr;
		}

	        newPtr = malloc(sizeof(DomainNode));
        
       		if (newPtr != NULL) {  /* is space available? */
			memset(newPtr->name, 0, sizeof(newPtr->name));
			strncpy(newPtr->name, domain, 10);
			newPtr->data = data;
			newPtr->fileCount = 1;
			newPtr->seconds = secs;
			currentPtr = *sPtr;
			while (currentPtr->nextPtr != NULL)
				currentPtr = currentPtr->nextPtr;
			currentPtr->nextPtr = newPtr;
			newPtr->nextPtr = NULL;
#ifdef DEBUG
			printf("New domain structure for \"%s\" created.\n", newPtr->name);
#endif
			return 1;
		} else {
			printf("fatal: out of memory.\n");
			exit(1);
		}
	}	
	return 0;
} /* doDomain */


int compareDates(char date1[25], char date2[25])
{
	float dw1, dw2;
	int i, m = 0;
	char foo[5] = {0};

	if (strlen(date1) == 24) {
		strncpy(foo, date1 + 20, 4);
		dw1 = atoi(foo) * 4800;
		strncpy(foo, date2 + 20, 4);
		dw2 = atoi(foo) * 4800;
		strncpy(foo, date1 + 4, 3);
		foo[3] = '\0';
		for (i=0; i <= 11; i++) {
			if (!strcmp(MONTHS[i], foo)) {
				m = i;
				break;
			}
		}
#ifdef DEBUG
		if (!m) printf("ERROR: invalid month\n");
#endif	
		if (m) dw1 += m * 400;

		m = 0;
		strncpy(foo, date2 + 4, 3);
		for (i=0; i <= 11; i++) {
			if (!strcmp(MONTHS[i], foo)) {
				m = i;
				break;
			}
		}
#ifdef DEBUG
		if (!m) printf("ERROR: invalid month\n");
#endif	
		if (m) dw2 += m * 400;

		strncpy(foo, date1 + 8, 2);
		foo[3] = '\0';
		dw1 += atoi(foo);
		strncpy(foo, date2 + 8, 2);
		dw2 += atoi(foo);
	} else {
		strncpy(foo, date1 + 16, 4);
		dw1 = atoi(foo) * 4800;
		strncpy(foo, date2 + 16, 4);
		dw2 = atoi(foo) * 4800;
		strncpy(foo, date1, 3);
		foo[3] = '\0';
		for (i=0; i <= 11; i++) {
			if (!strcmp(MONTHS[i], foo)) {
				m = i;
				break;
			}
		}
#ifdef DEBUG
		if (!m) printf("ERROR: invalid month\n");
#endif	
		if (m) dw1 += m * 400;

		m = 0;
		strncpy(foo, date2, 3);
		for (i=1; i <= 12; i++) {
			if (!strcmp(MONTHS[i], foo)) {
				m = i;
				break;
			}
		}
#ifdef DEBUG
		if (!m) printf("ERROR: invalid month\n");
#endif	
		if (m) dw2 += m * 400;

		strncpy(foo, date1 + 4, 2);
		foo[3] = '\0';
		dw1 += atoi(foo);
		strncpy(foo, date2 + 4, 2);
		dw2 += atoi(foo);
	}
	if (dw1 == dw2) return 0;
	if (dw1 > dw2)
		return 1;
	else
		return -1;
}	


void displayTotals(DOMPTR *sPtr, DAYPTR *dPtr, DIRPTR *drPtr, FILEPTR *fPtr)
{
	DOMPTR currentPtr, hiDPtr = NULL;
	DAYPTR currentDayPtr;
	DIRPTR currentDirPtr, hiPtr = NULL;
	FILEPTR currentFilePtr, hiFPtr = NULL;
	int i = 0, whee = 0;
	char date1[16] = {0}, date2[16] = {0};
	unsigned long long hibyte;

	strncpy(date1, lowDate, 10);
	strncat(date1, lowDate + 19, 5);
	strncpy(date2, highDate, 10);
	strncat(date2, highDate + 19, 5);

	printf("TOTALS FOR SUMMARY PERIOD %s TO %s\n\n", date1, date2);
	printf("Files Transmitted During Summary Period     %11u\n", fileCount);
	printf("Bytes Transmitted During Summary Period     %11llu\n\n", total);
	currentDayPtr = *dPtr;
	for (; currentDayPtr != NULL; i++)
		currentDayPtr = currentDayPtr->nextPtr;
	printf("Average Bytes Transmitted Per File          %11llu\n", total / fileCount);
	printf("Average Files Transmitted Daily             %11u\n", fileCount / i);
	printf("Average Bytes Transmitted Daily             %11llu\n\n", total / i);
	if (!numberDailyStats || numberDailyStats >= i) printf("Daily Transmission Statistics\n\n");
	else printf("Daily Transmission Statistics (last %d days)\n\n", numberDailyStats);
	printf("                 Number Of    Number of    Average    Percent Of  Percent Of\n");
	printf("     Date        Files Sent  Bytes  Sent  Xmit  Rate  Files Sent  Bytes Sent\n");
	printf("---------------  ----------  -----------  ----------  ----------  ----------\n");
	currentDayPtr = *dPtr;
	while(currentDayPtr != NULL) {
		whee++;
		if (!numberDailyStats || whee > i - numberDailyStats) {
			printf("%s  %10u  %11llu  %5.1f KB/s  %10.2f  %10.2f\n", currentDayPtr->name, currentDayPtr->fileCount, currentDayPtr->data,
				(float) currentDayPtr->data / currentDayPtr->seconds / 1024, (float) currentDayPtr->fileCount / fileCount * 100, (float) currentDayPtr->data / total * 100);
		}
		currentDayPtr = currentDayPtr->nextPtr;
	}
	if (totalTraffic) {
		if (numberSectionStats) printf("%c\nTop %d Total Transfers from each Archive Section (By bytes)\n\n", 12, numberSectionStats);
		else printf("%c\nTotal Transfers from each Archive Section (By bytes)\n\n", 12);
		printf("                                                                 ---- %%  Of ----\n");
		printf("                 Archive Section                Files Bytes Sent  Files   Bytes\n");
   		printf(" ---------------------------------------------- ----- ---------- ------- -------\n");
		currentDirPtr = *drPtr;
		for (i = whee = 0; currentDirPtr != NULL; i++)
			currentDirPtr = currentDirPtr->nextPtr;
		while (whee < i && (!numberSectionStats || whee < numberSectionStats)) {
			hibyte = 0;
			currentDirPtr = *drPtr;
			while(currentDirPtr != NULL) {
				if (!currentDirPtr->sorted && currentDirPtr->data >= hibyte) {
					hiPtr = currentDirPtr;
					hibyte = currentDirPtr->data;
				}
				currentDirPtr = currentDirPtr->nextPtr;
			}
			whee++;
			hiPtr->sorted = 1;
			printf(" %-46s %5u %10llu %7.2f %7.2f\n", hiPtr->name, hiPtr->fileCount, hibyte, (float) hiPtr->fileCount / fileCount * 100, (float) hibyte / total * 100);
		}
	}
	if (domainTraffic) {
		if (!numberDomainStats) printf("%c\nTotal Transfer Amount By Domain (By bytes)\n\n", 12);
		else printf("%c\nTop %d Total Transfer Amounts By Domain (By bytes)\n\n", 12, numberDomainStats);
		printf("             Number Of    Number of     Average    Percent Of  Percent Of\n");
		printf("Domain Name  Files Sent   Bytes Sent   Xmit  Rate  Files Sent  Bytes Sent\n");
		printf("-----------  ----------  ------------  ----------  ----------  ----------\n");
		currentPtr = *sPtr;
		for (i = whee = 0; currentPtr != NULL; i++)
			currentPtr = currentPtr->nextPtr;
		while (whee < i && (!numberDomainStats || whee < numberDomainStats)) {
			hibyte = 0;
			currentPtr = *sPtr;
			while (currentPtr != NULL) {
				if (!currentPtr->sorted && currentPtr->data > hibyte) {
					hiDPtr = currentPtr;
					hibyte = currentPtr->data;
				}
				currentPtr = currentPtr->nextPtr;
			}
			if (!hibyte) break;
			whee++;
			hiDPtr->sorted = 1;
			printf("%-11s  %10u  %12llu  %5.1f KB/s  %10.2f  %10.2f\n", hiDPtr->name, hiDPtr->fileCount, hibyte,
				(float) hibyte / hiDPtr->seconds / 1024, (float) hiDPtr->fileCount / fileCount * 100, (float) hibyte / total * 100);
		}
	}
	printf("\nThese figures only reflect FTP transfers.  There are many sites\n");
	printf("which mount the archives via NFS, and those transfers are not logged\n"); 
	printf("and reported by this program.\n\n");
	if (hourlyTraffic) {
		printf("%c\nHourly Transmission Statistics\n\n", 12);
		printf("                Number Of    Number of     Average    Percent Of  Percent Of\n");
		printf("     Time       Files Sent  Bytes  Sent  Xmit   Rate  Files Sent  Bytes Sent\n");
		printf("--------------  ----------  -----------  -----------  ----------  ----------\n");
		for (i = 0; i <= 23; i++)
			if (hourStruct[i].fileCount)
				printf("%-14d  %10u  %11llu  %6.2f KB/s  %10.2f  %10.2f\n",
					i, hourStruct[i].fileCount, hourStruct[i].data, (float) hourStruct[i].data / hourStruct[i].seconds / 1024, (float) hourStruct[i].fileCount / fileCount * 100, (float) hourStruct[i].data / total * 100);
	}
	if (mostDownloads) {
		printf("%c\nTop %d Most Downloaded Files (By number)\n\n", 12, mostDLNumber);
		printf("                                                                ---- %%  Of ----\n");
		printf("                           Filename                       Files  Files   Bytes\n");
		printf(" -------------------------------------------------------- ----- ------- -------\n");
		currentFilePtr = *fPtr;
		for (i = whee = 0; currentFilePtr != NULL; i++)
			currentFilePtr = currentFilePtr->nextPtr;
		while (whee < mostDLNumber) {
			hibyte = 0;
			currentFilePtr = *fPtr; 
			while(currentFilePtr != NULL) {
				if (!currentFilePtr->sorted && currentFilePtr->fileCount >= hibyte) {
					hiFPtr = currentFilePtr;
					hibyte = currentFilePtr->fileCount;
				}
				currentFilePtr = currentFilePtr->nextPtr;
			}
			if (!hibyte) break;
			whee++;
			hiFPtr->sorted = 1;
			if (strlen(hiFPtr->name) > 56) {
				char tempName[55];
				strcpy(tempName, hiFPtr->name + strlen(hiFPtr->name) - 54); 
				printf("...%-54s %5llu %7.2f %7.2f\n", tempName, hibyte, (float) hibyte / fileCount * 100, (float) hiFPtr->data / total * 100);
			} else
				printf(" %-56s %5llu %7.2f %7.2f\n", hiFPtr->name, hibyte, (float) hibyte / fileCount * 100, (float) hiFPtr->data / total * 100);
		}
	}
} /* displayTotals */


void displayHTMLTotals(DOMPTR *sPtr, DAYPTR *dPtr, DIRPTR *drPtr, FILEPTR *fPtr)
{
	DOMPTR currentPtr, hiDPtr = NULL;
	DAYPTR currentDayPtr;
	DIRPTR currentDirPtr, hiPtr = NULL;
	FILEPTR currentFilePtr, hiFPtr = NULL;
	int i = 0, whee = 0;
	char date1[16] = {0}, date2[16] = {0};
	float hixfer = 0;
	unsigned long long hibyte;
#ifndef NOGRAPH
	gdImagePtr image1, image2, image3;
	int black, black2, black3, white, white2, white3, blue, blue2, blue3,
		green, green2, green3;
	char foo[5], tempPath[1024];
	float hixfer2 = 0;
#endif

	strncpy(date1, lowDate, 10);
	strncat(date1, lowDate + 19, 5);
	strncpy(date2, highDate, 10);
	strncat(date2, highDate + 19, 5);

	printf("<HTML><HEAD>\n");
	printf("<TITLE>xferstats: %s to %s</TITLE>\n", date1, date2);
	printf("</HEAD><BODY TEXT=\"#FFFFFF\" BGCOLOR=\"#000000\">\n");

	printf("<H1>TOTALS FOR SUMMARY PERIOD %s TO %s</H1><br>\n", date1, date2);
	printf("<TABLE BORDER = 2>\n");
	printf("<TR><TH ALIGN=LEFT>Files Transmitted During Summary Period</TH><TH ALIGN=RIGHT>%11u</TH></TR>\n", fileCount);
	printf("<TR><TH ALIGN=LEFT>Bytes Transmitted During Summary Period</TH><TH ALIGN=RIGHT>%11llu</TH></TR>\n", total);
	currentDayPtr = *dPtr;
	for (; currentDayPtr != NULL; i++)
		currentDayPtr = currentDayPtr->nextPtr;
	printf("<TR><TH ALIGN=LEFT>Average Bytes Transmitted Per File</TH><TH ALIGN=RIGHT>%11llu</TH></TR>\n", total / fileCount);
	printf("<TR><TH ALIGN=LEFT>Average Files Transmitted Daily</TH><TH ALIGN=RIGHT>%11u</TH><TR>\n", fileCount / i);
	printf("<TR><TH ALIGN=LEFT>Average Bytes Transmitted Daily</TH><TH ALIGN=RIGHT>%11llu</TH></TR></TABLE><BR><BR>\n", total / i);
	if (!numberDailyStats || numberDailyStats >= i) printf("<H1>Daily Transmission Statistics</H1><BR>\n");
	else printf("<H1>Daily Transmission Statistics (last %d days)</H1><BR>\n", numberDailyStats);
	printf("<TABLE BORDER=2>\n");
	printf("<TR><TH></TH><TH>Number Of</TH><TH>Number of</TH><TH>Average</TH><TH>Percent Of</TH><TH>Percent Of</TH></TR>\n");
	printf("<TR><TH>Date</TH><TH>Files Sent</TH><TH>Bytes Sent</TH><TH>Xmit Rate</TH><TH>Files Sent</TH><TH>Bytes Sent</TH></TR>\n");
	currentDayPtr = *dPtr;
	while(currentDayPtr != NULL) {
		whee++;
		if (!numberDailyStats || whee > i - numberDailyStats) {
			printf("<TR><TH>%s</TH><TH ALIGN=RIGHT>%u</TH><TH ALIGN=RIGHT>%llu</TH><TH ALIGN=RIGHT>%5.1f KB/s</TH><TH ALIGN=RIGHT>%10.2f</TH><TH ALIGN=RIGHT>%10.2f</TH></TR>\n", currentDayPtr->name, currentDayPtr->fileCount, currentDayPtr->data,
				(float) currentDayPtr->data / currentDayPtr->seconds / 1024, (float) currentDayPtr->fileCount / fileCount * 100, (float) currentDayPtr->data / total * 100);
		}
		currentDayPtr = currentDayPtr->nextPtr;
	}
	printf("</TABLE><BR><BR>\n");
	if (totalTraffic) {
		if (numberSectionStats) printf("<H1>Top %d Total Transfers from each Archive Section (By bytes)</H1><BR><BR>\n", numberSectionStats);
		else printf("<H1>Total Transfers from each Archive Section (By bytes)</H1><BR><BR>\n");
		printf("<TABLE BORDER=2>\n");
		printf("<TR><TH>Archive Section</TH><TH>Files</TH><TH>Bytes Sent</TH><TH>%% of Files</TH><TH>%% of Bytes</TH></TR>\n");
		currentDirPtr = *drPtr;
		for (i = 0; currentDirPtr != NULL; i++)
			currentDirPtr = currentDirPtr->nextPtr;
		whee = 0;
		while (whee < i && (!numberSectionStats || whee < numberSectionStats)) {
			hibyte = 0;
			currentDirPtr = *drPtr;
			while(currentDirPtr != NULL) {
				if (!currentDirPtr->sorted && currentDirPtr->data >= hibyte) {
					hiPtr = currentDirPtr;
					hibyte = currentDirPtr->data;
				}
				currentDirPtr = currentDirPtr->nextPtr;
			}
			whee++;
			hiPtr->sorted = 1;
			printf("<TR><TH ALIGN=LEFT>%s</TH><TH ALIGN=RIGHT>%u</TH><TH ALIGN=RIGHT>%llu</TH><TH ALIGN=RIGHT>%7.2f</TH><TH ALIGN=RIGHT>%7.2f</TH></TR>\n",
			hiPtr->name, hiPtr->fileCount, hibyte, (float) hiPtr->fileCount / fileCount * 100, (float) hibyte / total * 100);
		}
		printf("</TABLE><BR<BR>\n");
	}
	if (domainTraffic) {
		if (!numberDomainStats) printf("<H1>Total Transfer Amount By Domain (By bytes)</H1><BR>\n");
		else printf("<H1>Top %d Total Transfer Amounts By Domain (By bytes)</H1><BR>\n", numberDomainStats);
		printf("<TABLE BORDER=2>\n");
		printf("<TR><TH></TH><TH>Number Of</TH><TH>Number of</TH><TH>Average</TH><TH>Percent Of</TH><TH>Percent Of</TH></TR>\n");
		printf("<TR><TH>Domain Name</TH><TH>Files Sent</TH><TH>Bytes Sent</TH><TH>Xmit Rate</TH><TH>Files Sent</TH><TH>Bytes Sent</TH></TR>\n");
		currentPtr = *sPtr;
		for (i = whee = 0; currentPtr != NULL; i++)
			currentPtr = currentPtr->nextPtr;
		while (whee < i && (!numberDomainStats || whee < numberDomainStats)) {
			hibyte = 0;
			currentPtr = *sPtr;
			while (currentPtr != NULL) {
				if (!currentPtr->sorted && currentPtr->data > hibyte) {
					hiDPtr = currentPtr;
					hibyte = currentPtr->data;
				}
				currentPtr = currentPtr->nextPtr;
			}
			if (!hibyte) break;
			whee++;
			hiDPtr->sorted = 1;
			printf("<TR><TH>%s</TH><TH ALIGN=RIGHT>%u</TH><TH ALIGN=RIGHT>%llu</TH><TH ALIGN=RIGHT>%5.1f KB/s</TH><TH ALIGN=RIGHT>%10.2f</TH><TH ALIGN=RIGHT>%10.2f</TH></TR>\n", hiDPtr->name, hiDPtr->fileCount, hibyte,
				(float) hibyte / hiDPtr->seconds / 1024, (float) hiDPtr->fileCount / fileCount * 100, (float) hibyte / total * 100);
		}


		while(currentPtr != NULL) {
			currentPtr = currentPtr->nextPtr;
		}
		printf("</TABLE><BR><BR>\n");
	}
	printf("These figures only reflect FTP transfers.  There are many sites\n");
	printf("which mount the archives via NFS, and those transfers are not logged\n"); 
	printf("and reported by this program.<BR><BR>\n");
	if (hourlyTraffic) {
#ifndef NOGRAPH
		image1 = gdImageCreate(280, 280);
		image2 = gdImageCreate(280, 280);
		image3 = gdImageCreate(280, 280);
		black = gdImageColorAllocate(image1, 0, 0, 0);
		white = gdImageColorAllocate(image1, 255, 255, 255);
		blue = gdImageColorAllocate(image1, 135, 206, 235);
		green = gdImageColorAllocate(image1, 46, 139, 87);
		black2 = gdImageColorAllocate(image2, 0, 0, 0);
		white2 = gdImageColorAllocate(image2, 255, 255, 255);
		blue2 = gdImageColorAllocate(image2, 135, 206, 235);
		green2 = gdImageColorAllocate(image2, 46, 139, 87);
		black3 = gdImageColorAllocate(image3, 0, 0, 0);
		white3 = gdImageColorAllocate(image3, 255, 255, 255);
		blue3 = gdImageColorAllocate(image3, 135, 206, 235);
		green3 = gdImageColorAllocate(image3, 46, 139, 87);
		gdImageLine(image1, 30, 15, 30, 249, white);
		gdImageLine(image1, 30, 249, 270, 249, white);
		gdImageLine(image2, 30, 15, 30, 249, white2);
		gdImageLine(image2, 30, 249, 270, 249, white2);
		gdImageLine(image3, 30, 15, 30, 249, white3);
		gdImageLine(image3, 30, 249, 270, 249, white3);
		gdImageString(image1, gdFontTiny, 32, 250, "0   2   4   6   8", white);
		gdImageString(image1, gdFontTiny, 131, 250, "10  12  14  16  18  20  22", white);
		gdImageString(image2, gdFontTiny, 32, 250, "0   2   4   6   8", white2);
		gdImageString(image2, gdFontTiny, 131, 250, "10  12  14  16  18  20  22", white2);
		gdImageString(image3, gdFontTiny, 32, 250, "0   2   4   6   8", white3);
		gdImageString(image3, gdFontTiny, 131, 250, "10  12  14  16  18  20  22", white3);
		gdImageString(image1, gdFontTiny, 155 - 2 * gdFontTiny->w, 260, "Hour", white);
		gdImageString(image2, gdFontTiny, 155 - 2 * gdFontTiny->w, 260, "Hour", white2);
		gdImageString(image3, gdFontTiny, 155 - 2 * gdFontTiny->w, 260, "Hour", white3);
		gdImageStringUp(image1, gdFontTiny, 15, 140 + 12 * gdFontTiny->w, "Percent of Total Traffic", white);
		gdImageStringUp(image2, gdFontTiny, 15, 140 + 13 * gdFontTiny->w, "Average Throughput (in K/s)", white2);
		gdImageStringUp(image3, gdFontTiny, 15, 140 + 13 * gdFontTiny->w, "Maximum Throughput (in K/s)", white3);
#endif
		printf("<H1>Hourly Transmission Statistics</H1><BR>\n");
		printf("<IMG SRC=\"%s%s\"><IMG SRC=\"%s%s\"><BR>", HTML_GRAPH_PATH, HOURLY_DATA_GRAPH, HTML_GRAPH_PATH, HOURLY_THROUGHPUT_GRAPH);
		printf("<IMG SRC=\"%s%s\">", HTML_GRAPH_PATH, HOURLY_MAX_THROUGHPUT_GRAPH);
		printf("<TABLE BORDER=2>\n");
		printf("<TR><TH></TH><TH>Number Of</TH><TH>Number of</TH><TH>Average</TH><TH>Percent Of</TH><TH>Percent Of</TH></TR>\n");
		printf("<TR><TH>Time</TH><TH>Files Sent</TH><TH>Bytes Sent</TH><TH>Xmit Rate</TH><TH>Files Sent</TH><TH>Bytes Sent</TH></TR>\n");
		hibyte = 0;
		for (i = 0; i <= 23; i++) {
			if (hourStruct[i].data > hibyte)
				hibyte = hourStruct[i].data;
			if ((float) hourStruct[i].data / hourStruct[i].seconds / 1024 > hixfer)
				hixfer = (float) hourStruct[i].data / hourStruct[i].seconds / 1024;
#ifndef NOGRAPH
			if (hourStruct[i].hixfer > hixfer2)
				hixfer2 = hourStruct[i].hixfer;
#endif
		}
		for (i = 0; i <= 23; i++)
			if (hourStruct[i].fileCount) {
				printf("<TR><TH>%d</TH><TH ALIGN=RIGHT>%u</TH><TH ALIGN=RIGHT>%llu</TH><TH ALIGN=RIGHT>%6.2f KB/s</TH><TH ALIGN=RIGHT>%10.2f</TH><TH ALIGN=RIGHT>%10.2f</TH></TR>\n",
					i, hourStruct[i].fileCount, hourStruct[i].data, (float) hourStruct[i].data / hourStruct[i].seconds / 1024, (float) hourStruct[i].fileCount / fileCount * 100, (float) hourStruct[i].data / total * 100);
#ifndef NOGRAPH
				if (fmod(i, 2)) {
					gdImageFilledRectangle(image1, 10*i+31, (int) (248 - (float) hourStruct[i].data * 233 / hibyte), 10*(i+1)+30, 248, blue);
					gdImageFilledRectangle(image2, 10*i+31, (int) (248 - (float) hourStruct[i].data * 233 / hourStruct[i].seconds / 1024 / hixfer), 10*(i+1)+30, 248, blue2);
					gdImageFilledRectangle(image3, 10*i+31, (int) (248 - (float) hourStruct[i].hixfer * 233 / hixfer2), 10*(i+1)+30, 248, blue2);
				} else {
					gdImageFilledRectangle(image1, 10*i+31, (int) (248 - (float) hourStruct[i].data * 233 / hibyte), 10*(i+1)+30, 248, green);
					gdImageFilledRectangle(image2, 10*i+31, (int) (248 - (float) hourStruct[i].data * 233 / hourStruct[i].seconds / 1024 / hixfer), 10*(i+1)+30, 248, green2);
					gdImageFilledRectangle(image3, 10*i+31, (int) (248 - (float) hourStruct[i].hixfer * 233 / hixfer2), 10*(i+1)+30, 248, green2);
				}
#endif
			}
		printf("</TABLE><BR><BR>\n");
#ifndef NOGRAPH
		if (total) {
			sprintf(foo, "%3.1f", ((float) hibyte * 100 / total));
			gdImageString(image1, gdFontTiny, 29 - gdFontTiny->w * strlen(foo), 15, foo, white);
			sprintf(foo, "%3.1f", (float) hixfer);
 			gdImageString(image2, gdFontTiny, 29 - gdFontTiny->w * strlen(foo), 15, foo, white2);
			sprintf(foo, "%3.1f", (float) hixfer2);
 			gdImageString(image3, gdFontTiny, 29 - gdFontTiny->w * strlen(foo), 15, foo, white3);
		}
		memset(tempPath, 0, 1024);
		strncpy(tempPath, REAL_GRAPH_PATH, 1024);
		strcat(tempPath, HOURLY_DATA_GRAPH);
		gif = fopen(tempPath, "wb");
		gdImageGif(image1, gif);
		fclose(gif);
		gdImageDestroy(image1);
		memset(tempPath, 0, 1024);
		strncpy(tempPath, REAL_GRAPH_PATH, 1024);
		strcat(tempPath, HOURLY_THROUGHPUT_GRAPH);
		gif = fopen(tempPath, "wb");
		gdImageGif(image2, gif);
		fclose(gif);
		gdImageDestroy(image2);
		memset(tempPath, 0, 1024);
		strncpy(tempPath, REAL_GRAPH_PATH, 1024);
		strcat(tempPath, HOURLY_MAX_THROUGHPUT_GRAPH);
		gif = fopen(tempPath, "wb");
		gdImageGif(image3, gif);
		fclose(gif);
		gdImageDestroy(image3);
#endif
	}
	if (mostDownloads) {
		whee = 0;
		printf("<H1>Top %d Most Downloaded Files (by number)</H1><BR>\n", mostDLNumber);
		printf("<TABLE BORDER=2>\n");
		printf("<TR><TH></TH><TH>Number Of</TH><TH>Percent Of</TH><TH>Percent Of</TH></TR>\n");
		printf("<TR><TH>Filename</TH><TH>Files Sent</TH><TH>Files Sent</TH><TH>Bytes Sent</TH></TR>\n");
		currentFilePtr = *fPtr;
		for (i = 0; currentFilePtr != NULL; i++)
			currentFilePtr = currentFilePtr->nextPtr;
		while (whee < mostDLNumber) {
			hibyte = 0;
			currentFilePtr = *fPtr; 
			while(currentFilePtr != NULL) {
				if (!currentFilePtr->sorted && currentFilePtr->fileCount >= hibyte) {
					hiFPtr = currentFilePtr;
					hibyte = currentFilePtr->fileCount;
				}
				currentFilePtr = currentFilePtr->nextPtr;
			}
			if (hibyte == 0)
				break;
			whee++;
			hiFPtr->sorted = 1;
			if (strlen(hiFPtr->name) > 5000) {
				char tempName[55];
				strcpy(tempName, hiFPtr->name + strlen(hiFPtr->name) - 54); 
				printf("...%-54s %5llu %7.2f %7.2f\n", tempName, hibyte, (float) hibyte / fileCount * 100, (float) hiFPtr->data / total * 100);
			} else
				printf("<TR><TH ALIGN=LEFT>%s</TH><TH ALIGN=RIGHT>%llu</TH><TH ALIGN=RIGHT>%7.2f</TH><TH ALIGN=RIGHT>%7.2f</TH></TR>\n", hiFPtr->name, hibyte, (float) hibyte / fileCount * 100, (float) hiFPtr->data / total * 100);
		}
		printf("</TABLE>\n");
	}
} /* displayHTMLTotals */


void usage(void)
{
	fprintf(stderr, "%s\n", VERSION);
	fprintf(stderr, "usage: xferstats <options>\n\n"
			" -			grab the logfile from stdin\n"
			" -f <filename>		use <filename> for the log file\n"
			" -T			toggle logfile type (wuftpd or ncftpd)\n"
			" -H			toggle HTML output\n"
			" -r			include real users\n"
			" -a			include anonymous users\n"
			" -h			include report on hourly traffic\n"
			" -d [number]		include report on domain traffic\n"
			" -t [number]		report on total traffic by section\n" 
			" -m [number]		include a \"most downloaded files\" report\n"
			" -L <number>		limit the report-by-day report to [number] listings\n"
			" -A			include all users, all reports\n"
			" -D <domain>		report only on traffic from <domain>\n"
			" -l <number>		depth of path detail for sections (default 2)\n"
			" -s <section>		section to report on, for example: -s /pub will\n"
			"			report only on paths under /pub\n"
			" -v			display the version number\n");

} /* usage */


void parse_cmdline(int argc, char *argv[])
{
	int i = 1, j, realArg = 0;
	char c;

	for (; i < argc; i++) {
		if (*argv[i] == '-') {     /* argv[i] contains options */
			j = 1;
			if (*(argv[i] + 1) == '\0') use_stdin = 1;
			while ((c = *(argv[i] + j)) != '\0') {
				switch (c) {
					case 'f':
						if (i < argc - 1 && *argv[i + 1] != '-')
							fileStr = argv[i + 1];
						else {
							fprintf(stderr, "fatal: the -f parameter requires a filename following it.\n");
							exit(1);
						}
						break;
					case 'T':
						if (ftpd == 0) ftpd = 1;
						else ftpd = 0;
						break;
					case 'H':
						if (outputHTML) outputHTML = 0;
						else outputHTML = 1;
						break;
					case 'r':
						realArg = 1;
						realTraffic = 2;
						break;
					case 'a':
						realArg = 1;
						anonTraffic = 2;
						break;
					case 'h':
						realArg = 1;
						hourlyTraffic = 2;
						break;
					case 'd':
						realArg = 1;
						domainTraffic = 2;
						if (i < argc - 1 && *argv[i + 1] != '-') 
							if ((numberDomainStats = atoi(argv[i + 1])) < 0) {
								fprintf(stderr, "fatal: the -d parameter accepts only an optional number following it.\n");
								exit(1);
							}
						break;
					case 't':
						realArg = 1;
						totalTraffic = 2;
						if (i < argc - 1 && *argv[i + 1] != '-')
							if ((numberSectionStats = atoi(argv[i + 1])) < 0) {
								fprintf(stderr, "fatal: the -t parameter accepts only an optional number following it.\n");
								exit(1);
							}
						break;
					case 'm':
						realArg = 1;
						mostDownloads = 2;
						if (i < argc - 1 && *argv[i+1] != '-')
							if ((mostDLNumber = atoi(argv[i + 1])) < 1) {
								fprintf(stderr, "fatal: the -m parameter accepts only an optional number following it.\n");
								exit(1);
							}
						break;
					case 'L':
						if (i < argc - 1 && *argv[i + 1] != '-') {
							if ((numberDailyStats = atoi(argv[i + 1])) < 0) {
								fprintf(stderr, "fatal: the -L parameter requires a number greater than or equal to zero following it.\n");
								exit(1);
							}
						} else {
							fprintf(stderr, "fatal: the -L parameter requires a number following it.\n");
							exit(1);
						}
						break;
					case 'D':
						if (i < argc - 1 && *argv[i+1] != '-')
							strncpy(onlyDomain, argv[i+1], 4);
						else {
							fprintf(stderr, "fatal: the -D parameter requires a top-level domain following it.\n");
							exit(1);
						}
						printf("Transfer totals include the '%s' domain only.\n", onlyDomain);
						printf("All other domains are filtered out for this report.\n\n");
						break;
					case 'l':
						if (i < argc  - 1 && *argv[i+1] != '-')
							if ((depth = atoi(argv[i + 1]) < 1))
								fprintf(stderr, "fatal: the -l parameter accepts only a number greater than zero following it.\n");
						else {
							fprintf(stderr, "fatal: the -l parameter requires a number following it.\n");
							exit(1);
						}
						break;
					case 's':
						sectionTraffic = 1;
						if (i < argc - 1 && *argv[i+1] != '-')
							if (*argv[i+1] == '/') {
								strcpy(section, "/");
								strcat(section, argv[i + 1]);
							} else
								strncpy(section, argv[i + 1], 254);
						else {
							fprintf(stderr, "fatal: the -s parameter requires a path following it.\n");
							exit(1);
						}
						sectionLength = strlen(section);
						printf("Transfer totals include the '%s' section only.\n", section);
						printf("All other sections are filtered out for this report.\n\n");
						break;
					case 'v':
						fprintf(stderr, "%s\n", VERSION);
						exit(1);
					case 'A':
						realArg = 1;
#ifdef DEFAULT_ANON
						anonTraffic = 2;
#endif
#ifdef DEFAULT_HOURLY
						hourlyTraffic = 2;
#endif
#ifdef DEFAULT_DOMAIN
						domainTraffic = 2;
#endif
#ifdef DEFAULT_TOTALS
						totalTraffic = 2;
#endif
#ifdef DEFAULT_REAL
						realTraffic = 2;
#endif
#ifdef DEFAULT_MOSTDL
						mostDownloads = 2;
#endif
						break;
					default:
						usage();
						exit(1);
				}
				j++;
			}
		}
	}
	if (realArg) {
		if (anonTraffic < 2) anonTraffic = 0;
		if (hourlyTraffic < 2) hourlyTraffic = 0;
		if (domainTraffic < 2) domainTraffic = 0;
		if (totalTraffic < 2) totalTraffic = 0;
		if (realTraffic < 2) realTraffic = 0;
		if (mostDownloads < 2) mostDownloads = 0;
	}
} /* parse_cmdline */


void main(int argc, char **argv)
{
	DOMPTR startPtr = NULL;
	DAYPTR firstDayPtr = NULL;
	DIRPTR firstDirPtr = NULL;
	FILEPTR firstFilePtr = NULL;

	setbuf(stdout, NULL);
	
	fileStr = FILENAME;
	parse_cmdline(argc, argv);

	if (use_stdin) filePtr = stdin;
	else filePtr = fopen(fileStr, "r");
        if (filePtr == NULL) {
		perror(fileStr);
		exit(EXIT_FAILURE);
        }
	do {
		if (parseline()) {
			char c[11] = {0}, *d;

#ifdef DEBUG
			printf("%s ftped %s\n", host, fileName);
#endif
			if (strchr(host, '.')) {
				strncpy(c, strrchr(host, '.') + 1, 4);
#ifdef DEBUG
				printf("Last letter of domain: %d:%c\n", c[strlen(c)-1], c[strlen(c)-1]);
#endif
				if (isdigit(c[strlen(c)-1])) strcpy(c, "unresolved");
			} else
				strcpy(c, "unresolved");
			for(d = c; (*d = tolower(*d)); d++);
			strncpy(domain, c, 10);
			domain[10] = '\0';
			if (((!strcmp(onlyDomain, "") || !strcmp(onlyDomain, domain))) && (!sectionTraffic || !strncmp(fileName, section, sectionLength))) {
				total += data;
				fileCount++;
				if ((realTraffic && userType == 'r') || (anonTraffic && (userType == 'a' || userType == 'g'))) {
					if (!doDomain(&startPtr)) {
						printf("fatal: could not doDomain for domain \"%s\".\n", domain);
						exit(2);
					}
					if (!doDaily(&firstDayPtr)) {
						printf("fatal: could not doDaily.\n");
						exit(2);
					}
					if (!doDirs(&firstDirPtr, &firstFilePtr)) {
						printf("fatal: could not doDirs.\n");
						exit(2);
					}
					if (hourlyTraffic) doHours();
					if (!strcmp(lowDate, "")) {
						strncpy(lowDate, date, 24);
						lowDate[24] = '\0';
					} else if (compareDates(date, lowDate) < 0) {
						strncpy(lowDate, date, 24);
						lowDate[24] = '\0';
					}
					if (!strcmp(highDate, "")) {
						strncpy(highDate, date, 24);
						highDate[24] = '\0';
					} else if (compareDates(date, highDate) > 0) {
						strncpy(highDate, date, 24);
						highDate[24] = '\0';
					}
				}
			}
		}
	} while (!feof(filePtr));
	fclose(filePtr);
	if (startPtr != NULL)
		if (outputHTML)
			displayHTMLTotals(&startPtr, &firstDayPtr, &firstDirPtr, &firstFilePtr);
		else
			displayTotals(&startPtr, &firstDayPtr, &firstDirPtr, &firstFilePtr);
	else
		printf("There was no data to process.\n");
}
